@@ -14,6 +14,9 @@
 #include <glib/gi18n.h>
 #include <gtk/gtk.h>
 #include <unistd.h>
+#include <gudev/gudev.h>
+
+#include <sane/sane.h> // For SANE_STATUS_CANCELLED
 
 #include "ui.h"
 #include "scanner.h"
@@ -30,24 +33,12 @@ static Book *book;
 
 static gboolean scanning = FALSE;
 
-static gboolean clear_pages = FALSE;
-
 static gboolean first_autodetect = TRUE;
 
 
-static void
-scanner_ready_cb (Scanner *scanner)
-{
-    scanning = FALSE;
-    ui_set_scanning (ui, FALSE);
-}
-
-
 static void
 update_scan_devices_cb (Scanner *scanner, GList *devices)
 {
-    GList *dev_iter;
-
     if (first_autodetect) {
         first_autodetect = FALSE;
 
@@ -66,14 +57,8 @@ update_scan_devices_cb (Scanner *scanner, GList *devices)
         }
     }
 
-    /* Mark existing values as undetected */
-    ui_mark_devices_undetected (ui);
-
     /* Add/update detected devices */
-    for (dev_iter = devices; dev_iter; dev_iter = dev_iter->next) {
-        ScanDevice *device = dev_iter->data;       
-        ui_add_scan_device (ui, device->name, device->label);
-    }
+    ui_set_scan_devices (ui, devices);
 }
 
 
@@ -88,35 +73,34 @@ authorize_cb (Scanner *scanner, const gchar *resource)
 }
 
 
-static void
-scanner_page_info_cb (Scanner *scanner, ScanPageInfo *info)
+static Page *
+append_page (gboolean replace)
 {
     Page *page;
     Orientation orientation = TOP_TO_BOTTOM;
     gboolean do_crop = FALSE;
     gchar *named_crop = NULL;
-    gint cx, cy, cw, ch;
+    gint width = 100, height = 100, dpi = 100, cx, cy, cw, ch;
 
-    g_debug ("Page is %d pixels wide, %d pixels high, %d bits per pixel",
-             info->width, info->height, info->depth);
-    
+    /* Copy info from previous page */
     page = book_get_page (book, -1);
     if (page) {
         orientation = page_get_orientation (page);
+        width = page_get_scan_width (page);
+        height = page_get_scan_height (page);
+        dpi = page_get_dpi (page);
         
         do_crop = page_has_crop (page);
         if (do_crop) {
             named_crop = page_get_named_crop (page);
             page_get_crop (page, &cx, &cy, &cw, &ch);
         }
-    }
+    }  
 
-    if (clear_pages) {
+    if (replace)
         book_clear (book);
-        clear_pages = FALSE;
-    }
 
-    page = book_append_page (book, info->width, info->height, info->dpi, orientation);
+    page = book_append_page (book, width, height, dpi, orientation);
     if (do_crop) {
         if (named_crop)  {
             page_set_named_crop (page, named_crop);
@@ -125,11 +109,31 @@ scanner_page_info_cb (Scanner *scanner, ScanPageInfo *info)
         else
             page_set_custom_crop (page, cw, ch);
         page_move_crop (page, cx, cy);
-    }
-    
+    }   
     ui_set_selected_page (ui, page);
-
     page_start (page);
+  
+    return page;
+}
+
+
+static void
+scanner_page_info_cb (Scanner *scanner, ScanPageInfo *info)
+{
+    Page *page;
+
+    g_debug ("Page is %d pixels wide, %d pixels high, %d bits per pixel",
+             info->width, info->height, info->depth);
+
+    page = book_get_page (book, -1);
+  
+    /* Add a new page */
+    if (page_get_scan_line (page) != 0) {
+        page = append_page (FALSE);
+    }
+
+    g_return_if_fail (page != NULL);
+    page_set_scan_area (page, info->width, info->height, info->dpi);
 }
 
 
@@ -144,12 +148,19 @@ scanner_line_cb (Scanner *scanner, ScanLine *line)
 
 
 static void
-scanner_image_done_cb (Scanner *scanner)
+scanner_page_done_cb (Scanner *scanner)
 {
     Page *page;
-    
     page = book_get_page (book, book_get_n_pages (book) - 1);
     page_finish (page);
+}
+
+
+static void
+scanner_document_done_cb (Scanner *scanner)
+{
+    scanning = FALSE;
+    ui_set_scanning (ui, FALSE);
     ui_set_have_scan (ui, TRUE);
 }
 
@@ -157,11 +168,26 @@ scanner_image_done_cb (Scanner *scanner)
 static void
 scanner_failed_cb (Scanner *scanner, GError *error)
 {
-    ui_show_error (ui,
-                   /* Title of error dialog when scan failed */
-                   _("Failed to scan"),
-                   error->message,
-		   TRUE);
+    Page *page;
+
+    page = book_get_page (book, book_get_n_pages (book) - 1);
+
+    /* Remove a failed page */
+    if (page_get_scan_line (page) == 0)
+        book_delete_page (book, page);
+    else
+        page_finish (page);
+
+    if (!g_error_matches (error, SCANNER_TYPE, SANE_STATUS_CANCELLED)) {
+        ui_show_error (ui,
+                       /* Title of error dialog when scan failed */
+                       _("Failed to scan"),
+                       error->message,
+                       TRUE);
+    }
+        
+    ui_set_scanning (ui, FALSE);
+    ui_set_have_scan (ui, TRUE);
 }
 
 
@@ -180,31 +206,31 @@ scan_cb (SimpleScan *ui, const gchar *device, const gchar *profile_name, gboolea
           _("Scanned Document.pdf") },
         { "photo", 400, SCAN_MODE_COLOR,
           /* Default name for JPEG documents */
-          _("Scanned Document.jpeg") },
+          _("Scanned Document.jpg") },
         { "raw", 800, SCAN_MODE_COLOR,
           /* Default name for PNG documents */
           _("Scanned Document.png") },
         { NULL, 75, SCAN_MODE_COLOR,
           /* Default name for JPEG documents */
-          _("Scanned Document.jpeg") }                
+          _("Scanned Document.jpg") }                
     };
+    Page *page;
     gint i;
 
     g_debug ("Requesting scan of type %s from device '%s'", profile_name, device);
 
-    scanning = TRUE;
-    ui_set_have_scan (ui, FALSE);
-    ui_set_scanning (ui, TRUE);
-    
     /* Find this profile */
     for (i = 0; profiles[i].name && strcmp (profiles[i].name, profile_name) != 0; i++);
 
-    if (replace)
-        clear_pages = TRUE;
+    if (!scanning)
+        page = append_page (replace);
+
+    scanning = TRUE;
+    ui_set_have_scan (ui, FALSE);
+    ui_set_scanning (ui, TRUE);
+ 
     ui_set_default_file_name (ui, profiles[i].file_name);
-    scanner_scan (scanner, device, NULL, profiles[i].dpi, profiles[i].mode, 8, continuous);
-    //scanner_scan (scanner, device, "Flatbed", 50, mode, 8, continuous);
-    //scanner_scan (scanner, device, "Automatic Document Feeder", 200, mode, 8, continuous);
+    scanner_scan (scanner, device, profiles[i].dpi, profiles[i].mode, 8, continuous);
 }
 
 
@@ -418,9 +444,18 @@ get_options (int argc, char **argv)
 }
 
 
-int
-main(int argc, char **argv)
+static void
+on_uevent (GUdevClient *client, const gchar *action, GUdevDevice *device)
 {
+    scanner_redetect (scanner);
+}
+
+int
+main (int argc, char **argv)
+{
+    GUdevClient *udev_client;
+    const char *udev_subsystems[] = { "usb", NULL };
+
     bindtextdomain (GETTEXT_PACKAGE, LOCALE_DIR);
     bind_textdomain_codeset (GETTEXT_PACKAGE, "UTF-8");
     textdomain (GETTEXT_PACKAGE);
@@ -429,7 +464,7 @@ main(int argc, char **argv)
     gtk_init (&argc, &argv);
 
     get_options (argc, argv);
-
+  
     ui = ui_new ();
     book = ui_get_book (ui);
     g_signal_connect (ui, "start-scan", G_CALLBACK (scan_cb), NULL);
@@ -439,14 +474,17 @@ main(int argc, char **argv)
     g_signal_connect (ui, "quit", G_CALLBACK (quit_cb), NULL);
 
     scanner = scanner_new ();
-    g_signal_connect (G_OBJECT (scanner), "ready", G_CALLBACK (scanner_ready_cb), NULL);
     g_signal_connect (G_OBJECT (scanner), "update-devices", G_CALLBACK (update_scan_devices_cb), NULL);
     g_signal_connect (G_OBJECT (scanner), "authorize", G_CALLBACK (authorize_cb), NULL);
     g_signal_connect (G_OBJECT (scanner), "got-page-info", G_CALLBACK (scanner_page_info_cb), NULL);
     g_signal_connect (G_OBJECT (scanner), "got-line", G_CALLBACK (scanner_line_cb), NULL);
-    g_signal_connect (G_OBJECT (scanner), "image-done", G_CALLBACK (scanner_image_done_cb), NULL);
+    g_signal_connect (G_OBJECT (scanner), "page-done", G_CALLBACK (scanner_page_done_cb), NULL);
+    g_signal_connect (G_OBJECT (scanner), "document-done", G_CALLBACK (scanner_document_done_cb), NULL);
     g_signal_connect (G_OBJECT (scanner), "scan-failed", G_CALLBACK (scanner_failed_cb), NULL);
 
+    udev_client = g_udev_client_new (udev_subsystems);
+    g_signal_connect (udev_client, "uevent", G_CALLBACK (on_uevent), NULL);
+
     if (default_device)
         ui_set_selected_device (ui, default_device);
 
